@@ -171,7 +171,7 @@ sub html {
 
 	$word = &$mapFn($word) if $mapFn;
 	if (defined $word) {
-	    $facets .= '<a href="' . $url->as_string . '">' . encode_entities($word) . " ($counter)</a><br/>\n";
+	    $facets .= '<p><a href="' . $url->as_string . '">' . encode_entities($word) . " ($counter)</a></p>\n";
 	}
     }
 
