@@ -8,10 +8,12 @@ use HTML::Entities;
 use HTTP::Request::Common; # for PUT
 use URI::Escape qw(uri_escape_utf8);
 use Compress::Zlib;
+use XML::LibXML;
 use Carp qw(confess);
 use Masterkey::Admin::Record;
 use Masterkey::Admin::ResultSet;
 use Masterkey::Admin::Parent;
+use Masterkey::Admin::Master;
 use Masterkey::Admin::Properties;
 
 
@@ -227,7 +229,7 @@ sub _listURI {
 sub _torusURI {
     my $this = shift();
     my($realm) = @_;
-    
+
     # If $realm is undefined, then nothing is appended
     if (defined $realm && !$this->{isAuth}) {
 	# Realm override only works on the primary Torus
@@ -333,7 +335,7 @@ sub _get {
 	$this->{torusVer} = $torusVer;
 	#warn "set Torus's version to '$torusVer'";
     }
-    
+
     return $res;
 }
 
@@ -418,7 +420,8 @@ sub add {
 	%extra = (identityId => $user->field("identityId"));
     }
 
-    my $xml = $this->_makeXML(undef, {}, 'worldId' => $worldId, %data, %extra);
+    $data{worldId} = $worldId if defined $worldId;
+    my $xml = $this->_makeXML(undef, {}, %data, %extra);
     my $uri = $this->recordsURI($realm, undef, $user, 1);
     my $res = $this->_post($uri, $xml);
 
@@ -591,29 +594,26 @@ sub _makeXML {
 
     my %allKeys = (%data, %$goneHash);
 
-    my $xml = <<__EOT__;
-<?xml version="1.0" encoding="UTF-8" standalone="no"?>
-<record type="$type">
-  <layer name="override">
-__EOT__
-    foreach my $key (sort keys %allKeys) {
-	my $val = encode_entities($data{$key});
-	# Protect against stupid dashes pasted in from MS-Word
-	if (exists $goneHash->{$key}) {
-	    $xml .= "    <" . encode_entities($key) . " gone='yes'/>\n";
-	} elsif ($key ne "" && defined $data{$key}) {
-	    $val =~ s/\&ndash;/-/g;
-	    $xml .= "    <" . encode_entities($key) . ">";
-	    $xml .= $val;
-	    $xml .= "</" . encode_entities($key) . ">\n";
-	}
-    }
-    $xml .= <<__EOT__;
-  </layer>
-</record>
-__EOT__
+    my $doc = XML::LibXML::Document->new('1.0', 'utf-8');
+    my $root = $doc->createElement("record");
+    $root->setAttribute(type => $type);
 
-    return $xml;
+    my $layer = $doc->createElement('layer');
+    $layer->setAttribute(name => 'override');
+    $root->appendChild($layer);
+
+    for my $name (sort keys %allKeys) {
+	my $tag = $doc->createElement($name);
+	if (exists $goneHash->{$name}) {
+	    $tag->setAttribute(gone => 'yes');
+	} else {
+	    $tag->appendTextNode($data{$name});
+	}
+	$layer->appendChild($tag);
+    }
+
+    $doc->setDocumentElement($root);
+    return $doc->toString(1);
 }
 
 # You may pass either a simple hash or a Record object containing values
@@ -766,6 +766,14 @@ sub _addParentByURI {
     return $id;
 }
 
+sub _addMasterByURI {
+    my $this = shift();
+    my($uri, $label, $matchQuery, $goneHash, %data) = @_;
+
+    $this->_logTorusOp("addMaster", uri=>$uri, label=>$label, matchQuery=>$matchQuery);
+    my $xml = _master_xml(undef, $label, $matchQuery, $goneHash, %data);
+    my $res = $this->_post($uri, $xml);
+}
 
 sub parent {
     my $this = shift();
@@ -790,6 +798,28 @@ sub parents {
 	    @{ $simple->{parent} });
 }
 
+sub master {
+    my $this = shift();
+    my ($realm, $id) = @_;
+
+    $this->_logTorusOp("master", realm=>$realm, id=>$id);
+    my $uri = $this->masterURI($realm, $id);
+    my $simple = $this->_getAsXML($uri);
+    return new Masterkey::Admin::Master($this, $simple);
+}
+
+
+sub masters {
+    my $this = shift();
+    my ($realm, $query) = @_;
+
+    $this->_logTorusOp("masters", realm=>$realm, query=>$query);
+    my $uri = $this->mastersURI($realm, $query);
+    my $simple = $this->_getAsXML($uri);
+
+    return (map { new Masterkey::Admin::Master($this, $_, $using_XML_LibXML_Simple) }
+	    @{ $simple->{record} });
+}
 
 sub addParent {
     my $this = shift();
@@ -799,6 +829,12 @@ sub addParent {
 				  $name, $priority, $refreshAfter, $url);
 }
 
+sub addMaster {
+    my $this = shift();
+    my($realm, $label, $matchQuery, $goneHash, %data) = @_;
+
+    return $this->_addMasterByURI($this->mastersURI($realm), $label, $matchQuery, $goneHash, %data);
+}
 
 sub updateParent {
     my $this = shift();
@@ -857,6 +893,81 @@ sub parentsURI {
     return $uri;
 }
 
+### The "master" code here is disturbingly similar to the "parent" code.
+
+sub updateMaster {
+    my $this = shift();
+    my($realm, $id, $label, $matchQuery, $goneHash, %data) = @_;
+
+    $this->_logTorusOp("updateMaster", realm=>$realm, id=>$id, label=>$label,
+		      matchQuery=>$matchQuery);
+
+    my $uri = $this->masterURI($realm, $id);
+    my $xml = _master_xml($id, $label, $matchQuery, %data);
+    $this->_put($uri, $xml);
+}
+
+
+### disturbingly similar to _makeXML()
+sub _master_xml {
+    my($id, $label, $matchQuery, $goneHash, %data) = @_;
+
+    my %allKeys = (%data, %$goneHash);
+
+    my $doc = XML::LibXML::Document->new('1.0', 'utf-8');
+    my $root = $doc->createElement("record");
+    $root->setAttribute(type => 'master');
+
+    my $layer = $doc->createElement('layer');
+    $layer->setAttribute(name => 'master-override');
+    $layer->setAttribute(label => $label);
+    $layer->setAttribute(matchQuery => $matchQuery);
+    $root->appendChild($layer);
+
+    for my $name (sort keys %allKeys) {
+	my $tag = $doc->createElement($name);
+	if (exists $goneHash->{$name}) {
+	    $tag->setAttribute(gone => 'yes');
+	} else {
+	    $tag->appendTextNode($data{$name});
+	}
+	$layer->appendChild($tag);
+    }
+
+    $doc->setDocumentElement($root);
+    return $doc->toString(1);
+}
+
+
+sub deleteMaster {
+    my $this = shift();
+    my($realm, $id) = @_;
+
+    $this->_logTorusOp("deleteMaster", realm=>$realm, id=>$id);
+    my $uri = $this->masterURI($realm, $id);
+    $this->_delete_by_uri($uri);
+}
+
+
+sub masterURI {
+    my $this = shift();
+    my ($realm, $id) = @_;
+
+    my $uri = $this->mastersURI($realm);
+    return "$uri$id/";
+}
+
+
+sub mastersURI {
+    my $this = shift();
+    my ($realm, $query) = @_;
+
+    my $uri = $this->ws();
+    $uri .= ".$realm" if defined $realm;
+    $uri .= "/masters/";
+    die "query not supported in masters()" if defined $query && $query ne "";
+    return $uri;
+}
 
 sub properties {
     my $this = shift();
@@ -938,8 +1049,15 @@ sub _sortedList {
 
     my @facetNames;
     # The odd "categories:\054" tells the Torus to split the categories field on commas.
-    push @facetNames, 'categories:\054', 'serviceProvider', 'opacVendor', 'openAccess'
-	if $this->admin()->configValue("showFacets");
+    if ($this->admin()->configValue("showFacets")) {
+	my $facets = $this->admin()->configValue("facetList");
+	foreach my $facet (split(/\|/, $facets)) {
+	    ### Nasty hard-coded knowledge of specified fields' structures
+	    $facet .= ':\054' if $facet eq 'categories';
+	    $facet .= ':\040' if $facet eq 'capabilities';
+	    push @facetNames, $facet;
+	}
+    }
     push @facetNames, "displayName"
 	if $this->admin()->configValue("showTagCloud");
     my $facetNames;
@@ -978,7 +1096,7 @@ sub generateCloud {
 	    $cloud->add_term($word, $count) if length($word) > 1;
 	}
     }
-    
+
     $cloud->generate('uri' => ".", 'levels'=> $admin->configValue("tagCloudLevels"));
     return $cloud;
 }
